; 
;
; AUTOMATON implementation by Nemeth Gabor/nemeth_g@freemail.hu (2004.09.16.)
;
;   See the include file for brief explanations
;
; 

    .486                      ; create 32 bit code
    .model flat, stdcall      ; 32 bit memory model
    option casemap :none      ; case sensitive

;   include files
;   ~~~~~~~~~~~~~
    include \masm32\include\windows.inc
    include \masm32\include\kernel32.inc
    include \masm32\include\masm32.inc
    include \masm32\macros\macros.asm
    include \masm32\include\debug.inc    
    include automat.inc

;   libraries
;   ~~~~~~~~~
    includelib \masm32\lib\kernel32.lib
    includelib \masm32\lib\masm32.lib
    includelib \masm32\lib\debug.lib

    EXTERN CAtm_EMPTY:ABS

.data
                CAtm_DefaultTransTable      dd 0

.CODE
; 
;
; Creates an CAutomat class dynamically
;
; -------------------------------------------------------------------------
; USES esi
;
; OUT:      esi:    address of CAutomat
; 
CAtm_Create     PROC USES eax,Typ:DWORD,StartState:DWORD,AccStates_PTR:DWORD,TransTable:DWORD
                  ASSUME esi:PTR CFSM
                  .if (Typ == AUTOMATON_FSM)
                    mov esi,FUNC(HeapAlloc,FUNC(GetProcessHeap),0,SIZEOF CFSM)
                    mov [esi].CAtm_Type,AUTOMATON_FSM
                    
                    mov eax,StartState
                    mov [esi].CAtm_State,eax

                    mov eax,AccStates_PTR
                    mov [esi].CAtm_AccStates_PTR,eax
                    mov eax,TransTable
                    mov [esi].CAtm_TransTable_PTR,eax

                    mov [esi].CAtm_NextState_PTR,offset CFSM_NextState
; -------------------------------------------------------------------------
                  .elseif (Typ == AUTOMATON_EFSM)
                  ASSUME esi:PTR CEFSM
                    mov esi,FUNC(HeapAlloc,FUNC(GetProcessHeap),0,SIZEOF CEFSM)
                    mov [esi].CAtm_Type,AUTOMATON_EFSM
                    mov eax,AccStates_PTR
                    mov [esi].CAtm_AccStates_PTR,eax
                    mov eax,TransTable
                    mov [esi].CAtm_TransTable_PTR,eax
                    mov [esi].CAtm_NextState_PTR,offset CEFSM_NextState
; -------------------------------------------------------------------------
                  .elseif (Typ == AUTOMATON_ESSM)
                  ASSUME esi:PTR CESSM
                    mov esi,FUNC(HeapAlloc,FUNC(GetProcessHeap),0,SIZEOF CESSM)
                    mov [esi].CAtm_Type,AUTOMATON_ESSM
                    mov eax,AccStates_PTR
                    mov [esi].CAtm_AccStates_PTR,eax
                    mov eax,TransTable
                    mov [esi].CAtm_TransTable_PTR,eax
                    mov eax,FUNC(HeapAlloc,FUNC(GetProcessHeap),0,65536)
                    mov [esi].CAtm_StackEnd_PTR,eax
;PrintHex eax,'::Stack End'
                    add eax,10000h
;PrintHex eax,'::Stack Start'
                    mov [esi].CAtm_StackStart_PTR,eax
                    mov [esi].CAtm_Stack_PTR,eax
                    mov [esi].CAtm_NextState_PTR,offset CESSM_NextState
                  .endif
                    ret
                  ASSUME esi:nothing
CAtm_Create     ENDP


; 
;
; Destroys the FSM
;
; 
CAtm_Destroy    PROC USES esi,AUTOMAT:DWORD
                  ASSUME esi:PTR CFSM
                    mov esi,AUTOMAT
                    invoke HeapFree,FUNC(GetProcessHeap),0,esi
                    ret
                  ASSUME esi:nothing
CAtm_Destroy    ENDP

; 
;
; Checks wheter the automaton is in an accept state?
;
; -------------------------------------------------------------------------
; OUT:      CF:     0 if accepted
;                   1 if rejected
; 
CAtm_Check      PROC USES eax ebx ecx esi,AUTOMAT:DWORD
                  ASSUME esi:PTR CFSM
                    mov esi,AUTOMAT

                  .if ([esi].CAtm_Type == AUTOMATON_FSM)
                    mov ebx,[esi].CAtm_AccStates_PTR    ;the set of accept states
                    mov ecx,[ebx]                       ;number of states in set
                    mov eax,[esi].CAtm_State            ;current state
@Check1:
                    add ebx,4
                    cmp eax,[ebx]                       ;check it!
                    jz @Found
                    dec ecx
                    jnz @Check1
; -------------------------------------------------------------------------
                  .elseif ([esi].CAtm_Type == AUTOMATON_EFSM || [esi].CAtm_Type == AUTOMATON_ESSM)
                  ASSUME esi:PTR CEFSM
                    mov ebx,[esi].CAtm_AccStates_PTR    ;the set of accept states
                    mov ecx,[ebx]                       ;number of states in set
                    mov eax,[esi].CAtm_TransTable_PTR   ;current state
@Check2:
                    add ebx,4
                    cmp eax,[ebx]                       ;check it!
                    jz @Found
                    dec ecx
                    jnz @Check2
                  .endif

                    clc
                    ret
@Found:
                    stc
                    ret

                  ASSUME esi:nothing
CAtm_Check      ENDP

; 
;
; One step, next state
;
; -------------------------------------------------------------------------
; IN:       esi:    CAutomat class
;           eax:    input code
;
; OUT:      CF:     0 on success (input code found)
;                   1 on fail (input code not found)
; 
CFSM_NextState  PROC USES ebx ecx edx
                  ASSUME esi:PTR CFSM
                    mov edx,[esi].CAtm_State                    ;current state
                    mov ebx,[esi].CAtm_TransTable_PTR           ;set of transitions
                    mov ecx,[ebx]                               ;number of members in set
                    test ecx,ecx
                    jz @NotFound
                    add ebx,4
                  ASSUME ebx:PTR STransition                    ;start of the table
@SearchState:
                    cmp edx,[ebx].CAtm_StateIn                  ;check the state
                    jz @SearchCode
                    add ebx,SIZEOF STransition                  ;jump to the row in transition table
                    dec ecx
                    jnz @SearchState
                    stc
                    ret
@SearchCode:
                    cmp eax,[ebx].CAtm_Input                    ;Check the input code
                    jz @FoundCode
                    add ebx,SIZEOF STransition
                    dec ecx
                    jnz @SearchState
@NotFound:
                    stc
                    ret
@FoundCode:
                    mov ecx,[ebx].CAtm_StateOut                 ;Next state
                    mov [esi].CAtm_State,ecx
                    clc
                    ret
                  ASSUME ebx:nothing
                  ASSUME esi:nothing
CFSM_NextState  ENDP

; 
;
; One step, next state
;
; -------------------------------------------------------------------------
; IN:       esi:    CEFSM class
;           eax:    input code
;
; OUT:      CF:     0 on success (input code found)
;                   1 on fail (input not accepted)
; 
CEFSM_NextState PROC USES ebx ecx edx
                  ASSUME esi:PTR CEFSM
                    mov ebx,[esi].CAtm_TransTable_PTR           ;set of transitions
                    mov edx,[ebx+4]                             ;default action proc
                    mov ecx,[ebx]                               ;number of members in set
                    test ecx,ecx
                    jz @DefAction
                    add ebx,8
                  ASSUME ebx:PTR STransition2                   ;start of the table
@SearchCode:
                    cmp eax,[ebx].CAtm_Input                    ;Check the input code
                    jz @FoundCode
                    add ebx,SIZEOF STransition2
                    dec ecx
                    jnz @SearchCode
@NotFound:
                    stc
                    ret
@FoundCode:
                    mov ecx,[ebx].CAtm_StateOut                 ;Next state
                    mov [esi].CAtm_TransTable_PTR,ecx
                    mov ebx,[ebx].CAtm_Action                   ;Action proc
                    cmp ebx,NULL                                ;is legal?
                    jz @DefAction
                    call ebx                                    ;execute it!
                    ret
@DefAction:
                    cmp edx,NULL                                ;Default action
                    jz @NoAction
                    call edx
                    ret
@NoAction:
                    clc
                    ret
                  ASSUME ebx:nothing
                  ASSUME esi:nothing
CEFSM_NextState ENDP
; 
;
; One step, next state
;
; -------------------------------------------------------------------------
; IN:       esi:    CEFSM class
;           eax:    input code
;
; OUT:      CF:     0 on success (input code found)
;                   1 on fail (input not accepted)
; 
CESSM_NextState PROC USES eax ebx ecx edx edi
                  ASSUME esi:PTR CESSM
                    mov edi,[esi].CAtm_TransTable_PTR           ;set of transitions
;PrintText '----------------'
;PrintHex edi,'::Input State'
                    mov edx,[edi+4]                             ;default action proc
                    mov ecx,[edi]                               ;number of members in set
                    test ecx,ecx
                    jz @DefAction
                    add edi,8
                    CAtm_POP ebx
                    jnc @StackNotEmpty
                    mov ebx,CAtm_EMPTY
@StackNotEmpty:
                  ASSUME edi:PTR STransition3                   ;start of the table

;PrintHex eax,'::Input Code'
;PrintHex ebx,'::Stack Top'

@SearchCode1:

                    cmp eax,[edi].CAtm_Input                    ;Check the input code
                    jnz @SearchCode11
                    cmp ebx,[edi].CAtm_StackIn                  ;Check the stack top value
                    jz @FoundCode
@SearchCode11:
                    add edi,SIZEOF STransition3
                    dec ecx
                    jnz @SearchCode1
                    stc
                    ret
@FoundCode:
                    mov ecx,[edi].CAtm_StateOut                 ;Next state
                    mov ebx,[edi].CAtm_StackOut

                    mov [esi].CAtm_TransTable_PTR,ecx
                    cmp ebx,CAtm_EMPTY                          ;Got empty input?
                    jz @NoPush
                    CAtm_PUSH ebx
@NoPush:

;PrintHex ecx,'::Output State'

                    mov ebx,[edi].CAtm_Action                   ;Action proc
                    cmp ebx,NULL                                ;is legal?
                    jz @DefAction
                    call ebx                                    ;execute it!
                    ret
@DefAction:
                    cmp edx,NULL                                ;Default action
                    jz @NoAction
                    call edx
                    ret
@NoAction:
                    clc
                    ret
                  ASSUME ebx:nothing
                  ASSUME esi:nothing
CESSM_NextState ENDP

END